
#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tree.h"
#include "errors.h"

#include "lex.h"
#include "ilog-tree.h"

#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include "ggc.h"

static void scan_digit PARAMS((void));
static size_t get_filesize PARAMS ((const char* fname));
static void lex_string PARAMS ((char quote));
static enum TOK lex_name PARAMS ((void));
static int next_char PARAMS ((void));
static char escape_char PARAMS ((char quote));

/* if inc_new_line is true lexer add lineno when encounts newline */
static bool inc_new_line;

const char *lex_buffer;
int lex_cur_pos;
int lex_buffer_len;
int lex_last_token_head_pos;
int lex_last_token_tail_pos;

enum TOK current_token = TOK_UNKNOWN;

lex_value_t lex_value;

#define NEXT_CHAR (++lex_cur_pos)
#define CUR_CHAR (lex_buffer[lex_cur_pos])

const char *
last_token_str( void )
{
  return ggc_alloc_string( lex_buffer+lex_last_token_head_pos, LEX_LAST_TOKEN_LEN );
}

/* fixme: crlf */
static int
next_char ()
{
  while ( 1 ) {
    int c;

    if ( lex_cur_pos >= lex_buffer_len )
      return EOF;

    c = CUR_CHAR;
	
    if ( c == ' ' || c == '\t' ) {
      goto next;
    } if ( c == '\n' ) {

      if ( inc_new_line ) {
	input_line++;
      }
      goto next;
    }

    return c;

  next:
    lex_cur_pos++;
  }  
}

struct ab_resword GTY(())
{
  struct tree_common common;
  enum TOK token;
};

static void 
scan_digit ()
{
  tree val;
  int dig_value = 0;

  while ( ISDIGIT(CUR_CHAR) ) {
    dig_value *= 10;
    dig_value += CUR_CHAR - '0';
    lex_cur_pos++;
  }

  lex_last_token_tail_pos = lex_cur_pos;

  val = build_int_2( dig_value, 0 );
  lex_value.tree_value = val;
}

static char
escape_char( char quote )
{
  char c;
  lex_cur_pos++;
  c = CUR_CHAR;

  switch ( c ) {
  case 'n':
    return '\n';

  case 'e':
    return '\e';
  }

  if ( c==quote ) {
    return quote;
  }

  warning("unkown escape seaquence '%c'.", c);
  return c;
}

static void
lex_string( char quote )
{
  int len = 0;
  int size = 16;
  char *buf = alloca( size+1 );
  
  lex_cur_pos++;

  while ( CUR_CHAR != quote ) {

    if ( CUR_CHAR == '\\' ) {
      buf[len] = escape_char( quote );
    } else {
      buf[len] = CUR_CHAR;
    }

    lex_cur_pos++;
    len++;

    if ( len > size ) {
      char *old_buf = buf;

      size *= 2;
      buf = alloca( size+1 );

      memmove( buf, old_buf, len );
    }
  }

  buf[len] = '\0';
  
  lex_last_token_tail_pos = ++lex_cur_pos;

  if ( quote == '"' ) {
    lex_value.tree_value = build_ilog_string( buf );
  } else {
    if ( len != 1 ) {
      warning("multi byte char is quoted by (').");

      lex_value.int_value = buf[0];
    }
  }
}

#define ILOG_ID_IS_RESWORD(id) (TREE_LANG_FLAG_0(id))
#define ILOG_RID_CODE(id) (((ilog_identifier*)id)->rid_code)

static enum TOK
lex_name( ) 
{
  int len = 256;
  char *buf = alloca( len+1 );
  int i=0;
  tree id;

  while ( 1 ) {
    int c = CUR_CHAR;

    if ( !ISDIGIT(c) && !ISALPHA(c) && c != '_' ) {
      break;
    }

    lex_cur_pos++;

    buf[ i++ ] = c;
    if ( i>len ) {
      char *tmp = alloca( len*2 + 1 );
      memmove( tmp, buf, len );
      buf = tmp;
      len *= 2;
    }
  }

  buf[i] = '\0';
  lex_last_token_tail_pos = lex_last_token_head_pos + i;

  id = get_identifier ( buf );

  if ( ILOG_ID_IS_RESWORD(id) ) {
    return ILOG_RID_CODE(id);
  }

  lex_value.tree_value = id;
  return TOK_IDENT;
}

#define RET_TOKEN(tok)  do \
 { enum TOK ttookk = tok; current_token = ttookk; return ttookk; }\
while(0);

enum TOK
next_token()
{
  int c;

  c = next_char();
  lex_last_token_head_pos = lex_cur_pos;

  if ( ISDIGIT(c) ) {
    scan_digit ();
    RET_TOKEN (TOK_DIGCONST);
  }

  switch ( c ) {
  case '{':
  case '}':
  case '[':
  case ']':
  case '(':
  case ')':
  case ';':
  case '+':
  case '-':
  case '*':
  case '/':
  case ',':
  case '.':
    lex_cur_pos++;
    lex_last_token_tail_pos = lex_cur_pos;
    RET_TOKEN(c);

  case '&':
    lex_cur_pos++;

    if ( CUR_CHAR == '&' ) {
      lex_cur_pos++;
      lex_last_token_tail_pos = lex_cur_pos;

      RET_TOKEN( TOK_LSHIFT );
    } else {
      lex_last_token_tail_pos = lex_cur_pos;
      RET_TOKEN( '&' );
    }

  case '<':
    lex_cur_pos++;

    if ( CUR_CHAR == '<' ) {
      lex_cur_pos++;
      lex_last_token_tail_pos = lex_cur_pos;

      RET_TOKEN( TOK_LSHIFT );
    }  else if ( CUR_CHAR == '=' ) {
      lex_cur_pos++;
      lex_last_token_tail_pos = lex_cur_pos;

      RET_TOKEN( TOK_LE );
    } else {
      lex_last_token_tail_pos = lex_cur_pos;
      RET_TOKEN( '<' );
    }

  case '>':
    lex_cur_pos++;

    if ( CUR_CHAR == '>' ) {
      lex_cur_pos++;
      lex_last_token_tail_pos = lex_cur_pos;

      RET_TOKEN( TOK_RSHIFT );
    }  else if ( CUR_CHAR == '=' ) {
      lex_cur_pos++;
      lex_last_token_tail_pos = lex_cur_pos;

      RET_TOKEN( TOK_GE );
    } else {

      lex_last_token_tail_pos = lex_cur_pos;
      RET_TOKEN( '<' );
    }

  case '=':
    lex_cur_pos++;

    if ( CUR_CHAR == '=' ) {
      lex_cur_pos++;
      lex_last_token_tail_pos = lex_cur_pos;

      RET_TOKEN( TOK_EQUAL );
    } else {

      lex_last_token_tail_pos = lex_cur_pos;
      RET_TOKEN( '=' );
    }

  case '!':
    lex_cur_pos++;

    if ( CUR_CHAR == '=' ) {
      lex_cur_pos++;
      lex_last_token_tail_pos = lex_cur_pos;

      RET_TOKEN( TOK_NE );
    } else {

      lex_last_token_tail_pos = lex_cur_pos;
      RET_TOKEN( '!' );
    }

  case ':':
    lex_cur_pos++;

    if ( CUR_CHAR == ':' ) {
      lex_cur_pos++;
      lex_last_token_tail_pos = lex_cur_pos;

      RET_TOKEN( TOK_SCOPE_COLON );
    } else {

      lex_last_token_tail_pos = lex_cur_pos;
      RET_TOKEN( ':' );
    }

  case '"':
    lex_string('"');
    RET_TOKEN (TOK_STRCONST);

  case EOF: 
    RET_TOKEN (TOK_EOF);

  default:
    RET_TOKEN (lex_name());
  }
  
  RET_TOKEN (TOK_UNKNOWN);
}

void 
set_lex_buffer( const char *buf )
{
  int progsize = strlen( buf );

  if ( lex_buffer ) 
    free( (char*)lex_buffer );

  lex_cur_pos = 0;
  lex_buffer_len = progsize;
  inc_new_line = false;

  lex_buffer = buf;
}

void
unset_lex_buffer( void )
{
  lex_buffer = NULL;
}

bool
init_lex_buffer(fname)
     const char *fname;
{
  int fsize = get_filesize (fname);
  FILE *input;
  char *bufp;

  if ( fsize < 0 ) {
    return false;
  }

  if ( lex_buffer ) 
    free( (char*)lex_buffer);

  lex_buffer = xmalloc (fsize);
  lex_cur_pos = 0;
  lex_buffer_len = fsize;

  input = fopen (fname, "r");

  if ( input == NULL ) {
    return false;
  }

  bufp = (char*)lex_buffer;

  while ( fsize > 0 ) {
    int size = fread ( (char*)bufp, 1, fsize, input);
    fsize -= size;
    bufp += size;
  }

  inc_new_line = true;

  fclose( input );

  return true;
}

static size_t
get_filesize(fname)
     const char *fname;
{
  struct stat buf;
  if ( (stat (fname, &buf)) == -1 ) {
    perror (fname);
    return -1;
  }

  return buf.st_size;
}


struct resword
{
  const char *word;
  enum TOK token;
};

static const struct resword reswords[] = 
{
  {"if",TOK_IF},
  {"else",TOK_ELSE},
  {"while",TOK_WHILE},
  {"for", TOK_FOR},
  {"namespace",TOK_NAMESPACE},
  /* {"nldef",TOK_NLDEF}, */
  {"var",TOK_VAR},
  {"const",TOK_CONST},
  {"void",TOK_VOID},
  {"function",TOK_FUNCTION},
  {"using",TOK_USING},
  /* {"defun", TOK_DEFUN}, */
  /* {"defvar", TOK_DEFVAR}, */
  {"return", TOK_RETURN},
};

#define N_reswords (sizeof reswords / sizeof (struct resword))

void
init_reswords ()
{
  unsigned int i;
  
  for ( i=0; i<N_reswords; i++ ) {
    tree id = get_identifier (reswords[i].word);
    ILOG_ID_IS_RESWORD(id) = 1;
    ILOG_RID_CODE(id) = reswords[i].token;
  }
}
